// ----------------------------- SMS Relay Controller -----------------------------------------------------------------------------------------------

// Designed for use with Haute Solutions MeGGaBoard 3G (ATMega2560 with SIM5320 on UART1)
// using the Haute Solutions LED Switch Shield
// www.haute-solutions.com

// ----------------------------- User Configuration Block -------------------------------------------------------------------

#define CONFIRM_CMDS 1                   // Set to 1 if all cmds (on, off, etc) should get confirmation SMS reply
#define BOOT_NOTIFY 1                    // Set to 1 to send a Message to the Master Caller each time the system boots
#define HEARTBEAT_TM "01,12:00"          // Send Heartbeat (Keepalive) Status MSG on 1st Day of every Month at Noon - set to "" for None

// ----------------------------- SMSRelay Program Code -------------------------------------------------------------------------

// Version Info
#define VERSION "V5.2"

// Persistant Storage Info (EEPROM Info): Character Data, NULL Terminated
// All Fields are 20 Chars max (Inc NULL Terminator)
// The values below are the Field Offsets into EEPROM memory for start of each string
// Currently (V1 EEPROM Layout) Uses 320 Bytes Total

#define EESIG           0                // EEPROM Data Valid Signature ("SMSRELAY EEPROM V1")
#define EEONTIME        1                // ON Alarm Time: YY/MM/DD,HH:MM
#define EEONSCHEDULER   2                // Phone Number of Caller who set ON Alarm (for SMS Confirm)
#define EEOFFTIME       3                // OFF Alarm Time: YY/MM/DD,HH:MM
#define EEOFFSCHEDULER  4                // Phone Number of Caller who set OFF Alarm (for SMS Confirm)
#define EEPWD           5                // Password for Self Authorization
#define EEAUTHCALLERS   6                // Start of Authorized Caller Phone Numbers Array

#define EEPROMSig "SMSRELAY EEPROM V1"    // EEPROM Valid Signature
#define EEFIELDSIZE     20                // Size of each field/string (inc Terminator)
#define MAXAUTH         10                // Max Number of Authorized Callers

// Pins we will be using on the GBoard/GBoardHS
#define RESET_SW_PIN A0
#define RELAY_CTRL_PIN A1
#define MANUAL_SW_PIN A2
#define STATUS_LED_PIN A4
#define GSM_LED_PIN A5

#define BUFFSIZE 300
#define NUMSIZE 20
#define TIMESIZE 15
#define PWDSIZE 20

#include <EEPROM.h>
#include <GSMSMS.h>                      // GSM SMS Library
GSMSMS gsmsms(MEGGABOARD3G);

#include <LEDSwitchShield.h>             // LEDSwitchShield Library
LEDSWITCHSHIELD LEDSwitchShield(0);      // Use Standard LEDSwitch Shield Layout "0" (MeGGaBoard 3G)
//LEDSWITCHSHIELD LEDSwitchShield(1);    // Use Alternate LEDSwitch Shield Layout "1" (EGGBoard 3G)

#define NUMCMDS 11
char *CmdArray[NUMCMDS] = {
  "ON",             // 1
  "OFF",            // 2
  "PWD: ",          // 3 (Restricted to Master User)
  "STATUS",         // 4
  "AUTHLIST",       // 5
  "RESET",          // 6 (Restricted to Master User)
  "SIMINFO",        // 7 (Undocumented Diagnostic)
  "MEINFO",         // 8 (Undocumented Diagnostic)
  "CELLINFO",       // 9 (Undocumented Diagnostic)
  "OPERATORS",      // 10 (Undocumented Diagnostic)
  "ATCMD: "         // 11 (Undocumented Diagnostic)
};

typedef struct {
  int led;
  unsigned long msdelay;
  unsigned long timeref;
  int blinks;
  int state;
} BLINKSTRUCT;
int RTBlinker(BLINKSTRUCT *blinkstruct, int p2 = -3, unsigned long msdelay = 0);
BLINKSTRUCT Status_LED;
BLINKSTRUCT GSM_LED;

char Buffer[BUFFSIZE];                   // Primary work Buffer
char ontime[TIMESIZE];
char offtime[TIMESIZE];
char timenow[TIMESIZE];
char tcheck[TIMESIZE];
char caller[NUMSIZE];
char onscheduler[NUMSIZE];
char offscheduler[NUMSIZE];
char pwd[PWDSIZE];
char Temp[200];

int AuthCallers=0;
int ClockValid=0;
int HBLen = 0;

int csq;
unsigned long timer;

void setup()
{
int SIMStat;

  Serial.begin(9600);
  
  // Relay Control Pin
  pinMode(RELAY_CTRL_PIN, OUTPUT);
  
  // Dignostic LEDS
  RTBlinker(&Status_LED, STATUS_LED_PIN, 500);    // Initialize Status LED for Blinking
  RTBlinker(&GSM_LED, GSM_LED_PIN, 500);          // Initialize GSM LED for Blinking
  
  // Manual Swiches
  pinMode(MANUAL_SW_PIN, INPUT);
  pinMode(RESET_SW_PIN, INPUT);
  
  Serial.print("Starting: SMS Relay ");
  Serial.print(VERSION);
  Serial.println("...");
 
  // Initialize GSM Chip and set Operating Parameters
  if(gsmsms.Init(9600) == 0){
    gsmsms.SetParams();
    gsmsms.SendCmd("AT+CNMI=2,1");              // Enable Unsolicited SMS Notification
    gsmsms.SendCmd("AT+AUTOCSQ=1,1");           // Enable Unsolicited CSQ Notification (whenever it changes)
    Serial.println("GSM Ready");
  } else {
    Serial.println("GSM Init Failed!");
  }

  Serial.print("SIMCOM Chipset: ");
  Serial.println(gsmsms.GetSimCom());
 
  Serial.print("EEPROM ");
  if(IsEEPROMValid()) Serial.println("Valid");
  else {
    InitializeEEPROM();
    Serial.println("Initialized");
  }
  
  Serial.print("On Alarm ");
  if(GetOnTime(ontime, onscheduler)) {
    sprintf(Buffer,"%s (%s)", ontime, onscheduler);
    Serial.println(Buffer);
  } else {
    ontime[0]='\0';
    onscheduler[0]='\0';
    Serial.println("Not Set");
  }

  Serial.print("Off Alarm ");
  if(GetOffTime(offtime, offscheduler)) {
    sprintf(Buffer,"%s (%s)", offtime, offscheduler);
    Serial.println(Buffer);
  } else {
    offtime[0]='\0';
    offscheduler[0]='\0';
    Serial.println("Not Set");
  }

  Serial.print("Pwd: ");
  if(GetPwd(pwd)) {
    Serial.println(pwd);
  } else {
    pwd[0]='\0';
    Serial.println("Not Set");
  }

  gsmsms.DeleteAllSMS();

  tcheck[0]='\0';
  gsmsms.GetTime(timenow);
  HBLen = strlen(HEARTBEAT_TM);

  if(!gsmsms.SIMInserted()) {
    Serial.println("No SIM Inserted!");
    RTBlinker(&GSM_LED, -1);
    while (1) {
      RTBlinker(&GSM_LED);
    }
  }
  
  if((AuthCallers=EnumerateAuthCallers(1)) == 0) {
    Serial.println("No Auth Callers");
    RTBlinker(&Status_LED, -1);
  } else {
    RTBlinker(&Status_LED, -2);
    if(BOOT_NOTIFY) {
      GetAuthCaller(0, caller);
      SendStatus(caller,"System Ready!");
      SendAuthList(caller);
      Serial.println("Boot Notification Sent");
    }
  }

  timer=millis();
  GetCSQ();
  Serial.println("Ready!");
}

void loop() 
{
  // Service the Blinkers
  RTBlinker(&Status_LED);
  RTBlinker(&GSM_LED);

  if(GetTime()) {
    CheckAlarm(ontime, onscheduler, EEONTIME, EEONSCHEDULER, RELAY_CTRL_PIN, HIGH, "ON!");
    CheckAlarm(offtime, offscheduler, EEOFFTIME, EEOFFSCHEDULER, RELAY_CTRL_PIN, LOW, "OFF!");
    if(HBLen) {
      if(strncmp(timenow+6,HEARTBEAT_TM,HBLen) == 0) {
        GetAuthCaller(0, caller);
        SendStatus(caller,"HeartBeat!");
      }
    }
  } 
  
  if(CheckSwitch(MANUAL_SW_PIN)) {
    ToggleRelayState();
    Serial.println();
    if(digitalRead(RELAY_CTRL_PIN) == HIGH) Serial.println("ON!");
    else Serial.println("OFF!");
  }

  if(CheckSwitch(RESET_SW_PIN)) {
    FactoryReset();
    Serial.println();
    Serial.println("Reset");
  }

  ProcessUnsolicitedMsg();                  // Check for incoming SMS or CSQ change

  if(SerialMonitorInput()) {
    ProcessMonitorCommand();
    tcheck[0]='\0';                         // Force Redisplay of Time in Serial Monitor
  }

}

int GetTime() {
unsigned long tnow;

  tnow = millis();
  if (tnow < (timer + 30000)) return 0;   // Update the time every 30 secs to ensure we get 1 min resolution

  timer = tnow;
  gsmsms.GetTime(timenow);
  return 1;
}

void ProcessUnsolicitedMsg() {
int c;
int i;

  if(!gsmsms.Available()) return;

  Buffer[0]='\0';
  while(gsmsms.Available()) {
    c = gsmsms.Read(); 
    if ((i = strlen(Buffer)) < (BUFFSIZE - 1)) {
      Buffer[i] = c;  
      Buffer[i+1] = '\0';
    }
    if(!gsmsms.Available()) delay(100);
  }

  Serial.print("\n\n");
  Serial.println(Buffer);   // Local Echo

  if(strstr(Buffer,"+CMTI:")) {
    do {} while (CheckSMS() != 0);
    return;
  }

  if(strstr(Buffer,"+CSQ:")) {
    GetCSQ();
    return;
  }

  return;
}

void GetCSQ(){
int i;

  csq=gsmsms.GetSignalQuality();

  // Indicate/Update GSM Status (Based on CSQ)
  if(csq == 0) RTBlinker(&GSM_LED, 0);                  // OFF for poor signal
  else RTBlinker(&GSM_LED, -2);                         // ON if we have signal

  //  Indicate CSQ
  for(i=2; i<7; i++) LEDSwitchShield.LEDControl(1,i,0,0);  
  switch (csq) {
    case 5:
      LEDSwitchShield.LEDControl(1,2,1,0);
    case 4:
      LEDSwitchShield.LEDControl(1,3,1,0);
    case 3:
      LEDSwitchShield.LEDControl(1,4,1,0);
    case 2:
      LEDSwitchShield.LEDControl(1,5,1,0);
    case 1:
      LEDSwitchShield.LEDControl(1,6,1,0);
  }
}

void IndicateSMSIn() {
  LEDSwitchShield.ColumnDown(2,1,250);
  LEDSwitchShield.ColumnDown(2,0,250);
}

void IndicateSMSOut() {
  LEDSwitchShield.ColumnUp(2,1,250);
  LEDSwitchShield.ColumnUp(2,0,250);
}

int CheckSwitch(int swpin)
{
  if (digitalRead(swpin) == HIGH) {
    delay(500); // Debounce
    return 1;
  } else return 0;
}

void ToggleRelayState()
{
  if(digitalRead(RELAY_CTRL_PIN) == LOW) digitalWrite(RELAY_CTRL_PIN, HIGH);
  else digitalWrite(RELAY_CTRL_PIN, LOW);  
}

void FactoryReset()
{
int i;

  digitalWrite(STATUS_LED_PIN,LOW);
  
  InitializeEEPROM();
  gsmsms.DeleteAllSMS();
  
  ontime[0]='\0';
  onscheduler[0]='\0';
  offtime[0]='\0';
  offscheduler[0]='\0';
  pwd[0]='\0';
  AuthCallers=0;
}

int SerialMonitorInput()
{
int i;
char c;

  if(!Serial.available()) return 0;

  Buffer[0]='\0';
  while(Serial.available()) {
    c = Serial.read(); 
    if ((i = strlen(Buffer)) < (BUFFSIZE - 1)) {
      Buffer[i] = c;  
      Buffer[i+1] = '\0';
    }
    if(!Serial.available()) delay(100);
  }
  return strlen(Buffer);
}

void ProcessMonitorCommand()
{
  Serial.println();
  Serial.print("Cmd: ");
  Serial.print(Buffer);
  
  if(gsmsms.SendCmd(Buffer) == 0) Serial.println(gsmsms.GSMCommBuff);
  else {
    Serial.println();
    Serial.println("ERROR");
  }
}

int EnumerateAuthCallers(int show)
{
int i;

  for(i=0; i<MAXAUTH; i++) {
    if(GetAuthCaller(i, caller)) {
      if(show) {
        Serial.print("Auth ");
        Serial.print(i+1);
        Serial.print(": ");
        Serial.println(caller);
      }
    } else return i;   
  }
  
  return i+1;
}

int CheckAlarm(char *atime, char *acaller, int eetimefield, int eecallerfield, int pin, int state, char *msg)
{
int i;
int j;
int tnow;
int alarm;
int expired;
char tmp[3];

  if(strlen(atime) == 0) return 0;  // No Alarm set

  for(i=0, expired=0; i<15; i+=3) {
    strncpy(tmp, timenow+i, 2);
    tnow=atoi(tmp);
    strncpy(tmp, atime+i, 2);
    alarm=atoi(tmp);
    
    if(alarm > tnow) return 0;  // Not yet...
    if(alarm < tnow) {          // Expired
      expired=1;
      break;
    }    
  }
  
  Serial.println();
  Serial.print("ALARM: ");
  Serial.print(msg);
  Serial.print(" (");
  Serial.print(acaller);
  Serial.print(")");
  
  if(expired) {

    Serial.println(" Expired!");

  } else {                    // Alarm Triggered

    Serial.println();
    digitalWrite(pin, state);    // Control the relay (ON / OFF)
    SendStatus(acaller, msg);   // Send SMS Message to Scheduler
    
  }

  // Clear the Alarm in Local Memory
  atime[0]='\0';
  acaller[0]='\0';
  
  // Clear the Alarm in EEPROM
  EEPROM.write( eetimefield*EEFIELDSIZE, '\0');
  EEPROM.write( eecallerfield*EEFIELDSIZE, '\0');
  
  return 1;  // Triggered
}

int CheckSMS()
{
int auth;
 
  if(GetSMSWithBlink(caller, Buffer) == 0) return 0;

  //Is caller authorized?
  if(AuthCallers == 0) {
    if(strlen(caller) > 4) {
      if(AuthorizeCaller(caller)) auth=1;
    }
  } else auth=CheckAuthorized(caller);

  if(auth) {
    ProcessCommand(caller,Buffer);
  } else {
    if(strcasecmp(Buffer,pwd) == 0) {
      if(AuthorizeCaller(caller)) auth=1;  
    } else {
      Serial.println();
      Serial.print(caller);
      Serial.print(" NOT Authorized! (");
      Serial.print(Buffer);
      Serial.println(")");
      Serial.print(timenow);
      ForwardMsg(caller);
      return -1;
    }
  }

  return 1;
}

int ForwardMsg(char *from)
{
int i;
int pushval;
int msglen;
int oldloc;
int newloc;

  pushval = (6 + strlen(from) + 2);
  Buffer[BUFFSIZE-pushval-1]='\0';
  msglen = strlen(Buffer);
  for(i=msglen; i>0; i--) {
    oldloc=i-1;
    newloc=oldloc+pushval;
    Buffer[newloc] = Buffer[oldloc];
  }
  strncpy(Buffer,"(From ",6);
  strncpy(Buffer+6,from,strlen(from));
  strncpy(Buffer+6+strlen(from),") ",2); 

  GetAuthCaller(0, caller);
  SendSMSWithBlink(caller,Buffer);
}

int AuthorizeCaller(char *caller)
{
char *authmsg="Authorized!";

  if(AuthCallers == MAXAUTH) {
    //  Send SMS stating max auth users?
    return 0;
  }
  
  WriteEEPROMField(EEAUTHCALLERS+AuthCallers, caller);
  if(AuthCallers == 0) RTBlinker(&Status_LED, -2);  // Turn on STATUS LED (Stop Flashing)
  AuthCallers++;
  
  Serial.println();
  Serial.print(caller);
  Serial.print(" ");
  Serial.print(authmsg);
  
  SendSMSWithBlink(caller,authmsg);    // Send SMS Message to Scheduler

  return AuthCallers;
}

int CheckAuthorized(char *caller)
{
int i;
char contents[20];

  for(i=0; i<MAXAUTH; i++) {
    ReadEEPROMField(EEAUTHCALLERS+i, contents);
    if(contents[0] == '\0') break;
    if(!strcmp(caller,contents)) return 1;  
  }
  
  return 0;
}

void ProcessCommand(char *caller, char *Buffer)
{
char *p1;
char *param;
char tmp[20];

  Serial.println();
  Serial.print("(");
  Serial.print(caller);
  Serial.print(") ");
  Serial.println(Buffer);

  param=strchr(Buffer,' ')+1;
  switch (GetCommandID(Buffer, CmdArray, NUMCMDS)) {
    case 0:   // Invalid Command
      SendSMSWithBlink(caller,"What?");
      return;

    case 1:   // ON
      if(!strncasecmp(Buffer,"ON @",4)) {
        ontime[0]='\0';
        if(strlen(Buffer) >= 19) {
          strncpy(ontime,Buffer+5,14);
          ontime[14]='\0';
        }
        strcpy(onscheduler,caller);
        SetOnTime(ontime, onscheduler);
        if(CONFIRM_CMDS) SendStatus(caller, "Scheduled!");
      } else {
        digitalWrite(RELAY_CTRL_PIN, HIGH);
        if(CONFIRM_CMDS) SendStatus(caller, "ON!");
      }
      return;
      
    case 2:   // OFF
      if(!strncasecmp(Buffer,"OFF @",5)) {
        offtime[0]='\0';
        if(strlen(Buffer) >= 20) {
          strncpy(offtime,Buffer+6,14);
          offtime[14]='\0';
        }
        strcpy(offscheduler,caller); 
        SetOffTime(offtime, offscheduler); 
        if(CONFIRM_CMDS) SendStatus(caller, "Scheduled!");
      } else {
        digitalWrite(RELAY_CTRL_PIN, LOW);
        if(CONFIRM_CMDS) SendStatus(caller, "OFF!");
      }
      return;

    case 3:   // PWD:
      ReadEEPROMField(EEAUTHCALLERS,tmp);
      if(strcmp(caller,tmp) == 0) {  // Only First Auth User Can Set PWD
        if(strlen(Buffer+5)) {
          WriteEEPROMField(EEPWD, Buffer+5);
          strcpy(pwd,Buffer+5);
        } else {
          EEPROM.write( EEPWD, '\0');
          strcpy(pwd,"");
        }
        SendSMSWithBlink(caller,"PWD Set!");       
      } else {
        SendSMSWithBlink(caller,"Not Authorized!");
      }
      return;
    
    case 4:   // STATUS
      SendStatus(caller, NULL);
      return;
      
    case 5:   // AUTHLIST
      SendAuthList(caller);
      return;
      
    case 6:   // RESET
      ReadEEPROMField(EEAUTHCALLERS,tmp);
      if(strcmp(caller,tmp) == 0) {  // Only First Auth User Can Set PWD
        FactoryReset();
        Serial.println();
        Serial.println("Reset");
        SendSMSWithBlink(caller,"Factory Reset!  Don't forget to re-authorize and set Password");       
      } else {
        SendSMSWithBlink(caller,"Not Authorized!");
      }
      return;
      
    case 7:   // SIMINFO (Undocumented)
      SendSIMInfo(caller);
      return;
      
    case 8:   // MEINFO (Undocumented)
      SendMEInfo(caller);
      return;
      
    case 9:   // CELLINFO (Undocumented)
      SendCellInfo(caller);
      return;
      
    case 10:   // OPERATORS (Undocumented)
      SendOperators(caller);
      return;
      
    case 11:   // ATCMD: (Undocumented)
      p1=strchr(param,'\n');
      if(p1 != NULL) *p1='\0';
      gsmsms.SendCmdParams(param,30000,1000,"OK",10);
      strncpy(Buffer,gsmsms.GSMCommBuff,BUFFSIZE);
      Buffer[BUFFSIZE-1]='\0';
      SendSMSWithBlink(caller,Buffer);
      return;
  }
}

int GetCommandID(char *astring, char *CmdArray[], int NumCmds) {
int i;
 
  for (i=0; i<NumCmds; i++) {
    if(!strncasecmp(astring,CmdArray[i],strlen(CmdArray[i]))) return i+1;
  }
  return 0;
}

void SendSIMInfo(char *acaller){
char *p1;
char *p2;
char *p3;

  strcpy(Buffer,"");

  // ICCID
  gsmsms.GetICCID(Temp);
  strcat(Buffer,"ICCID: ");
  strcat(Buffer,Temp);
  strcat(Buffer,"\n");
  
  // IMSI
  gsmsms.GetIMSI(Temp);
  strcat(Buffer,"IMSI: ");
  strcat(Buffer,Temp);
  strcat(Buffer,"\n");

  // Service Provider
  gsmsms.GetATResult("AT+CSPN?","+CSPN: ",Temp,1000);
  p1=strchr(Temp,'\"');
  if(p1 != NULL) {
    p1++;
    p2=strchr(p1,'\"');
    if(p2 != NULL) {
      *p2='\0';
      p3=p1;
    } else p3=Temp;
  } else p3=Temp;
  strcat(Buffer,"Provider: ");
  strcat(Buffer,p3);
  strcat(Buffer,"\n");
  
  // Subscriber Number
  gsmsms.GetSubscriberNumber(Temp);
  strcat(Buffer,"Number: ");
  strcat(Buffer,Temp);
  
  SendSMSWithBlink(acaller,Buffer);
}

void SendMEInfo(char *acaller) {
  
  strcpy(Buffer,"");

  // Chipset ID
  gsmsms.GetChipset(Temp);
  strcat(Buffer,"Chipset: ");
  strcat(Buffer,Temp);
  strcat(Buffer,"\n");
  
  // IMEI
  gsmsms.GetIMEI(Temp);
  strcat(Buffer,"IMEI: ");
  strcat(Buffer,Temp);

  SendSMSWithBlink(acaller,Buffer);  
}

void SendCellInfo(char *acaller) {
char *p1;
char *p2;
char *p3;

  strcpy(Buffer,"");

  // Current Band
  switch(gsmsms.GetSimCom()) {
    case 808:
    case 900:
      gsmsms.GetATResult("AT+CBAND?","+CBAND: ",Temp,1000);
      break;
    case 5320:
      gsmsms.GetATResult("AT*CNTI=0","*CNTI: 0,",Temp,1000);
      break;
  }
  strcat(Buffer,"Band: ");
  strcat(Buffer,Temp);
  strcat(Buffer,"\n");

  // Signal Quality
  sprintf(Temp,"Signal: %c/5",char(csq+48));
  strcat(Buffer,Temp);
  strcat(Buffer,"\n");
 
  // Current Network Operator and Mode
  gsmsms.GetATResult("AT+COPS?","+COPS: ",Temp,10000);
  p1=strchr(Temp,'\"');
  if(p1 != NULL) {
    p1++;
    p2=strchr(p1,'\"');
    if(p2 != NULL) {
      *p2='\0';
      p3=p1;
    } else p3=Temp;
  } else p3=Temp;
  strcat(Buffer,"Current: ");
  strcat(Buffer,p3);
  strcat(Buffer,"\n");
  
  // Service Center Address
  gsmsms.GetATResult("AT+CSCA?","+CSCA: ",Temp,1000);
  p1=strchr(Temp,'\"');
  if(p1 != NULL) {
    p1++;
    p2=strchr(p1,'\"');
    if(p2 != NULL) {
      *p2='\0';
      p3=p1;
    } else p3=Temp;
  } else p3=Temp;
  strcat(Buffer,"SMSC: ");
  strcat(Buffer,p3);
  strcat(Buffer,"\n");

  SendSMSWithBlink(acaller,Buffer);  
}
 
void SendOperators(char *acaller) {

  strcpy(Buffer,"");

  // Network Operators Currently in the Network
  gsmsms.GetATResult("AT+COPS=?","+COPS: ",Temp,60000);
  strcat(Buffer,"Operators: ");
  strcat(Buffer,Temp);
  strcat(Buffer,"\n");

  SendSMSWithBlink(acaller,Buffer);  
}

// ----------------------------- Routines which deal with Persistent (EEPROM Data) ---------------------------------------------

int GetOnTime(char *atime, char *caller) {
  if (ReadEEPROMField(EEONTIME, atime) == 0) return 0;
  if (ReadEEPROMField(EEONSCHEDULER, caller) == 0) return 0;
  return strlen(atime);
}

void SetOnTime(char *atime, char *caller) {
  WriteEEPROMField(EEONTIME, atime);
  WriteEEPROMField(EEONSCHEDULER, caller);
}

int GetOffTime(char *atime, char *caller) {
  if (ReadEEPROMField(EEOFFTIME, atime) == 0) return 0;
  if (ReadEEPROMField(EEOFFSCHEDULER, caller) == 0) return 0;
  return strlen(atime);
}

void SetOffTime(char *atime, char *caller) {
  WriteEEPROMField(EEOFFTIME, atime);
  WriteEEPROMField(EEOFFSCHEDULER, caller);
}

int GetPwd(char *pwd) {
  if (ReadEEPROMField(EEPWD, pwd) == 0) return 0;
  return strlen(pwd);
}

void SetPwd(char *pwd) {
  WriteEEPROMField(EEPWD, pwd);
}

int GetAuthCaller(int pos, char *caller) {
  if (ReadEEPROMField(EEAUTHCALLERS+pos, caller) == 0) return 0;
  return strlen(caller);
}

void SetAuthCaller(int pos, char *caller) {
  WriteEEPROMField(EEAUTHCALLERS+pos, caller);
}

int IsEEPROMValid() {

  if(ReadEEPROMField(EESIG,Buffer) < strlen(EEPROMSig)) return 0;
  if(strcmp(Buffer, EEPROMSig)) return 0;
  return 1;
}

void InitializeEEPROM() {
int i;
  
  WriteEEPROMField(EESIG, EEPROMSig);
  for(i=1; i<(EEAUTHCALLERS+MAXAUTH); i++) EEPROM.write( i*EEFIELDSIZE, '\0');
}

int WriteEEPROMField(int field, char *text) {
  return WriteEEPROMString( field*EEFIELDSIZE, text, EEFIELDSIZE);
}

int ReadEEPROMField(int field, char *text) {
  return ReadEEPROMString( field*EEFIELDSIZE, text, EEFIELDSIZE);
}

int WriteEEPROMString(int offset, char *text, int maxlen) {
int i;
int c;

  if((offset+maxlen) >= 512) return 0;

  for (i=0; i<maxlen; i++) {
    if(i == (maxlen-1)) EEPROM.write(offset+i, '\0');
    else EEPROM.write(offset+i, text[i]);
    if(text[i] == '\0') break;
  }
  
  return i+1;
}

int ReadEEPROMString(int offset, char *text, int maxlen) {
int i;

  if((offset+maxlen) >= 512) return 0;

  for (i=0; i<maxlen; i++) {
    text[i] = EEPROM.read(offset+i);
    if(text[i] == '\0') break;
  }
  
  text[maxlen-1]='\0';

  return i+1;
}

// ----------------------------- Routines which communicate directly with the "Inline" GSMSMS Library --------------------------

int GetSMSWithBlink(char *number, char *msg)
{
int stat;

  stat = gsmsms.GetSMS(number, msg, NUMSIZE, BUFFSIZE);
  if(stat > 0) IndicateSMSIn();
  return stat;
}

int SendSMSWithBlink(char *number, char *msg)
{
  IndicateSMSOut();
  return gsmsms.SendSMS(number, msg);
}

int SendAuthList(char *acaller)
{
int  i;
char msgend[2]={0x1a,'\0'};

  //  We are going to build our SMS Message incrementally using the low level library functions
  //  So we don't have to allocate a bunch of memory for a temp string

  if(gsmsms.PrepToSendSMS(acaller) > 0) return 1;
  
  gsmsms.Print("Authorized Callers List: ");
  gsmsms.Print("\n\n");
    
  for(i=0; i<MAXAUTH; i++) {
    if(GetAuthCaller(i, Temp)) {
      gsmsms.Print("Auth ");
      gsmsms.Print(i+1);
      gsmsms.Print(": ");
      gsmsms.Print(Temp);
      gsmsms.Print("\n");
    } else break;   
  }
  
  IndicateSMSOut();
  
  gsmsms.Print(msgend);
  if(gsmsms.FlushSMS() > 0) return 2;
    
  return 0;
}

int SendStatus(char *acaller, char *msg)
{
int  i;
int  sigstars;
char msgend[2]={0x1a,'\0'};
char csqval[2];

  //  We are going to build our SMS Message incrementally using the low level library functions
  //  So we don't have to allocate a bunch of memory for a temp string

  if(gsmsms.PrepToSendSMS(acaller) > 0) return 1;
  
  if(msg != NULL) {
    gsmsms.Print(msg);
    gsmsms.Print("\n\n");
  }

  gsmsms.Print("SMS Relay: ");
  gsmsms.Print(VERSION);
  gsmsms.Print("\n");
    
  gsmsms.Print("Time: ");
  gsmsms.Print(timenow);
  gsmsms.Print("\n");
    
  gsmsms.Print("Status: ");
  if(digitalRead(RELAY_CTRL_PIN) == HIGH) gsmsms.Print("ON");
  else gsmsms.Print("OFF");
  gsmsms.Print("\n");
  
  gsmsms.Print("On @ ");
  if(strlen(ontime) == 0) {
    gsmsms.Print("(Not Set)");
  } else {
    gsmsms.Print(ontime);
  }
  gsmsms.Print("\n");
  
 gsmsms.Print("Off @ ");
  if(strlen(offtime) == 0) {
    gsmsms.Print("(Not Set)");
  } else {
    gsmsms.Print(offtime);
  }
  gsmsms.Print("\n");

  csqval[0]=char(csq+48);
  csqval[1]='\0';
  gsmsms.Print("Signal: ");
  gsmsms.Print(csqval);
  gsmsms.Print("/5");
  gsmsms.Print("\n");
  
/*  
  for(i=0; i<MAXAUTH; i++) {
    if(GetAuthCaller(i, Temp)) {
      gsmsms.Print("Auth ");
      gsmsms.Print(i+1);
      gsmsms.Print(": ");
      gsmsms.Print(Temp);
      gsmsms.Print("\n");
    } else break;   
  }
*/  
  IndicateSMSOut();
  
  gsmsms.Print(msgend);
  if(gsmsms.FlushSMS() > 0) return 2;
    
  return 0;
}

// ----------------------------- RTBlinker Support --------------------------

int RTBlinker(BLINKSTRUCT *blinkstruct, int p2, unsigned long msdelay) {
int blinkcount;

// if msdelay is specified (!= 0) then p2=led pin number for initialization
// else p2=blinkcount for control

  if(msdelay == 0l) blinkcount=p2;            // Control Call
  else {                                      // Initialization Call
    blinkstruct->led = p2;
    if(blinkstruct->led >= 0) pinMode(blinkstruct->led, OUTPUT);
    blinkstruct->msdelay = msdelay;
    blinkstruct->timeref = 0;
    blinkstruct->blinks = 0;
    blinkstruct->state = -1;
    return 0;
  }

// blinkcount Control:
//
//  -3 = Continue Blinking / Processing
//  -2 = Solid On (No Blinking)
//  -1 = Set Continuous Blink (no blink pattern)
//   0 = Stop Blinking
//  1+ = Set Blink Pattern

// Return Codes:
//
// Low -> High = 1  (State Change)
// High -> Low = 2  (State Change)
// High = -1        (No Change)
// Low = -2         (No Change)
// Idle = 0
// Sequence Completed = 3

int elapsed;
  
  switch(blinkcount) {
    case -3:                  // Continue Blinking/Processing
      if(blinkstruct->state < 0) return 0;
      break;
    case -2:                   // Set Solid On
      blinkstruct->blinks = blinkcount;
      if(blinkstruct->led >= 0) digitalWrite(blinkstruct->led, HIGH);
      blinkstruct->timeref = 0;
      blinkstruct->state = 1;
      return 1;
    case 0:                   // Stop Blinking / Turn Off
      blinkstruct->blinks = blinkcount;
      if(blinkstruct->led >= 0) digitalWrite(blinkstruct->led, LOW);
      blinkstruct->timeref = 0;
      blinkstruct->state = -1;
      return 2;
    default:                  // Set Repeating Blink (>0) or Continuous Blink (-1)
      blinkstruct->blinks=blinkcount;
      blinkstruct->timeref = millis();
      blinkstruct->state=0;
      return 1;
  }

  if(blinkstruct->blinks == -2) return -1;

  elapsed = int(((millis()-blinkstruct->timeref))/blinkstruct->msdelay);      // Elapsed Time Units

  if((elapsed < (blinkstruct->blinks*2)) || (blinkstruct->blinks == -1)) {
    if (elapsed % 2) {                            // Odd (Off)
      if(blinkstruct->state == 1) {
        if(blinkstruct->led >= 0) digitalWrite(blinkstruct->led, LOW);
        blinkstruct->state=0;
        return 2;
      }
      return -1;
    } else {                                      // Even (On)
      if(blinkstruct->state == 0) {
        if(blinkstruct->led >= 0) digitalWrite(blinkstruct->led, HIGH);
        blinkstruct->state=1;
        return 1;
      }  
      return -2;
    }
  } else {
    if(blinkstruct->state == 1) {
      if(blinkstruct->led >= 0) digitalWrite(blinkstruct->led, LOW);
      blinkstruct->state=0;
      return 2;
    }
    if(elapsed >= ((blinkstruct->blinks*2)+2)) {
      blinkstruct->timeref=millis();
      if(blinkstruct->led >= 0) digitalWrite(blinkstruct->led, HIGH);
      blinkstruct->state=1;
      return 3;
    } 
    return -2;
  }
}
